/* Copyright 2007-2010, Unpublished Work of Technologic Systems
 * All Rights Reserved.
 *
 * THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
 * PROPRIETARY AND TRADE SECRET INFORMATION OF TECHNOLOGIC SYSTEMS.
 * ACCESS TO THIS WORK IS RESTRICTED TO (I) TECHNOLOGIC SYSTEMS 
 * EMPLOYEES WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE
 * OF THEIR ASSIGNMENTS AND (II) ENTITIES OTHER THAN TECHNOLOGIC
 * SYSTEMS WHO HAVE ENTERED INTO APPROPRIATE LICENSE AGREEMENTS.  NO
 * PART OF THIS WORK MAY BE USED, PRACTICED, PERFORMED, COPIED, 
 * DISTRIBUTED, REVISED, MODIFIED, TRANSLATED, ABRIDGED, CONDENSED, 
 * EXPANDED, COLLECTED, COMPILED, LINKED, RECAST, TRANSFORMED, ADAPTED
 * IN ANY FORM OR BY ANY MEANS, MANUAL, MECHANICAL, CHEMICAL, 
 * ELECTRICAL, ELECTRONIC, OPTICAL, BIOLOGICAL, OR OTHERWISE WITHOUT
 * THE PRIOR WRITTEN PERMISSION AND CONSENT OF TECHNOLOGIC SYSTEMS.
 * ANY USE OR EXPLOITATION OF THIS WORK WITHOUT THE PRIOR WRITTEN
 * CONSENT OF TECHNOLOGIC SYSTEMS COULD SUBJECT THE PERPETRATOR TO
 * CRIMINAL AND CIVIL LIABILITY.
 */
/* Modules:
 * Syscon -- DIO, LED, etc
 * CAN
 * XUARTs
 */
module ts7700_top(
  inout [15:0] smc_d_pad,
  input smc_clk_pad,
  input smc_cs1n_pad,
  input smc_llan_pad,
  input smc_luan_pad,
  input smc_be1n_pad,
  input smc_be2n_pad,
  input smc_wen_pad,
  input smc_rdn_pad,
  output smc_irq_pad,
  inout smc_rdy_pad,

  input fpga_25mhz_pad,
  output reboot_pad,
  output en_usb_5v_pad,
  output en_avdd_otgn_pad,
  output sd_powern_pad,
  output un_reset_pad,
  input fpga_resetn_pad,

  inout [54:0] dio_pad,
  output red_led_pad,
  output green_led_pad,

  inout sd_d0_pad,
  inout sd_d1_pad,
  inout sd_d2_pad,
  inout sd_d3_pad,
  inout sd_cmd_pad,
  output sd_clk0_pad,
  output sd_clk1_pad,
  output act_ledn_pad,

  input cpu_uart_txd_pad,
  input console_rxd_pad,
  inout console_txd_pad
);

parameter xuart_opt = 1'b1;
parameter can_opt = 1'b1;
parameter blockram_opt = 1'b1;
parameter sdcard_opt = 1'b1;

/****************************************************************************
 * TS-7700 boilerplate
 ****************************************************************************/

assign console_txd_pad = un_reset ? cpu_uart_txd_pad : 1'bz;

wire pll_100mhz, pll_locked, pll_75mhz, pll2_locked;
reg [15:0] rst_counter = 16'd0;
reg rstn = 1'b0;
reg lock_q;
always @(posedge wb_clk) begin
  lock_q <= pll_locked && pll2_locked && fpga_resetn_pad;
  if (lock_q) rst_counter <= rst_counter + 1'b1;
  else begin
    rst_counter <= 16'd0;
    rstn <= 1'b0;
  end
  if (rst_counter[15]) rstn <= 1'b1;
end
wire wb_rst = !rstn;
pll ts7700pll (
  .CLK(fpga_25mhz_pad),
  .CLKOP(pll_100mhz),
  .LOCK(pll_locked)
);
pll2 ts7700pll2 (
  .CLK(fpga_25mhz_pad),
  .CLKOP(pll_75mhz),
  .LOCK(pll2_locked)
);
/*
wire internal_osc;
reg [15:0] rst_counter = 16'd0;
reg rstn = 1'b0;
always @(posedge internal_osc) begin
  if (pll_locked && ) rst_counter <= rst_counter + 1'b1;
  else begin
    rst_counter <= 16'd0;
    rstn <= 1'b0;
  end
  if (rst_counter[15]) rstn <= 1'b1;
end
wire wb_rst = !rstn;
*/
wire reboot;
assign en_usb_5v_pad = 1'b1;
assign reboot_pad = reboot; 

reg en_avdd;
//assign en_avdd_otgn_pad = !en_avdd;
assign en_avdd_otgn_pad = 1'b0;
reg un_reset;
assign un_reset_pad = un_reset;

reg [31:0] usecs;
reg mode1, mode2;
wire [54:0] dio;
wire [54:0] dio_oe;
wire [54:0] dio_in;
always @(posedge wb_clk or posedge wb_rst) begin
  if (wb_rst) begin
    en_avdd <= 1'b0;
    un_reset <= 1'b0;
    mode1 <= 1'b0;
    mode2 <= 1'b0;
  end else begin
    en_avdd <= 1'b1;
    if (!un_reset) begin
      mode1 <= console_txd_pad;
      mode2 <= dio_in[30];
    end
    if (usecs[15]) un_reset <= 1'b1; // ~32 ms (16ms was not working)
  end
end

reg [6:0] count_1mhz;
always @(posedge wb_clk or posedge wb_rst) begin
  if (wb_rst) begin
    count_1mhz <= 7'd0;
    usecs <= 32'd0;
  end else begin
    count_1mhz <= count_1mhz + 1;
    //if (count_1mhz == 7'd99) begin
    if (count_1mhz == 7'd74) begin // change for 75MHz wb_clk
      count_1mhz <= 7'd0;
      usecs <= usecs + 1;
    end
  end
end

reg [2:0] count_12mhz; // actually 12.5 MHz
always @(posedge pll_100mhz or posedge wb_rst) begin
  if (wb_rst) count_12mhz <= 3'd0;
  else count_12mhz <= count_12mhz + 1'b1;
end

// pulling all SMC pins high during reset
//assign smc_d_pad[15:0] = un_reset ? {16{1'bz}} : {16{1'b1}};
//assign smc_d_pad[15:0] = {16{1'bz}}; 

/****************************************************************************
 * CPU bridge 
 ****************************************************************************/

wire smc_irq;
assign smc_irq_pad = smc_irq;

wire cpuwbm_cyc_o, cpuwbm_stb_o, cpuwbm_we_o;
wire [15:0] cpuwbm_dat_o;
wire [31:0] cpuwbm_adr, cpuwbm_adr_o, cpuwbm2_adr;
assign cpuwbm_adr_o = {15'd0, cpuwbm_adr[16:0]};
wire [14:0] ts4710_decodes = cpuwbm_adr[31:17];
wire [14:0] ts4710_decodes_100mhz = cpuwbm2_adr[31:17];
reg [15:0] cpuwbm_dat_i, cpuwbm2_dat_i;
wire [1:0] cpuwbm_sel_o;
reg cpuwbm_ack_i, cpuwbm2_ack_i;


wire [15:0] data_o;
wire data_oe;
assign smc_d_pad = data_oe ? data_o : 16'hzzzz;

wire smc_waitn;
assign smc_rdy_pad = smc_waitn;

/* Corrected notes on SMC_ADV# signal.  The built-in bootrom uses LLA#.  Once
 * we have control, we turn LLA# into a GPIO so that it is not used, and our
 * bus settings use #LUA.
 */

reg cpuwbm2_en;
wire cpuwbm2_cyc_o, cpuwbm2_stb_o, cpuwbm2_we_o;
wire [15:0] cpuwbm2_dat_o;
wire [31:0] cpuwbm2_adr_o, cpuwbs_adr_i;
assign cpuwbm2_adr_o = {15'd0, cpuwbm_adr[16:0]};
wire [1:0] cpuwbm2_sel_o;
wire [3:0] cpuwbm2_32_sel_o = 
  cpuwbm2_adr_o[1] ? {cpuwbm2_sel_o, 2'b00} : {2'b00, cpuwbm2_sel_o};
wire [15:0] cpuwbs_dat_i, cpuwbs_dat_o;
wire cpuwbs_cyc_i, cpuwbs_stb_i, cpuwbs_we_i, cpuwbs_ack_o;
wire [2:0] cpuwbs_cti_i;
wire [1:0] cpuwbs_bte_i;
mvsmcbridge cpu(
  .wb_rst_i(wb_rst),
  .wbm_clk_o(wb_clk),
  .wbm_cyc_o(cpuwbm_cyc_o),
  .wbm_stb_o(cpuwbm_stb_o),
  .wbm_dat_o(cpuwbm_dat_o),
  .wbm_adr_o(cpuwbm_adr),
  .wbm_we_o(cpuwbm_we_o),
  .wbm_dat_i(cpuwbm_dat_i),
  .wbm_ack_i(cpuwbm_ack_i),
  .wbm_sel_o(cpuwbm_sel_o),  

  /* WBM2 is at 100Mhz */
  .wbm2_en_i(cpuwbm2_en),
  .wbm2_clk_i(pll_100mhz),
  .wbm2_cyc_o(cpuwbm2_cyc_o),
  .wbm2_stb_o(cpuwbm2_stb_o),
  .wbm2_we_o(cpuwbm2_we_o),
  .wbm2_dat_o(cpuwbm2_dat_o),
  .wbm2_adr_o(cpuwbm2_adr),
  .wbm2_sel_o(cpuwbm2_sel_o),
  .wbm2_ack_i(cpuwbm2_ack_i),
  .wbm2_dat_i(cpuwbm2_dat_i),

  /* For satisfying read bursts */
  .wbs_clk_i(pll_100mhz),
  .wbs_cyc_i(cpuwbs_cyc_i),
  .wbs_stb_i(cpuwbs_stb_i),
  .wbs_we_i(cpuwbs_we_i),
  .wbs_adr_i(cpuwbs_adr_i),
  .wbs_dat_i(cpuwbs_dat_i),
  .wbs_dat_o(cpuwbs_dat_o),
  .wbs_cti_i(cpuwbs_cti_i),
  .wbs_bte_i(cpuwbs_bte_i),
  .wbs_ack_o(cpuwbs_ack_o),

  .smc_be1n_i(smc_be1n_pad),
  .smc_be2n_i(smc_be2n_pad),
  .smc_dat_i(smc_d_pad),
  .smc_dat_o(data_o),
  .smc_dat_oe_o(data_oe),
  .smc_rdn_i(smc_rdn_pad),
  .smc_wrn_i(smc_wen_pad),
  .smc_cs1n_i(smc_cs1n_pad),
  .smc_advn_i(smc_luan_pad & smc_llan_pad),
  .smc_waitn_o(smc_waitn),
  .smc_clk_i(smc_clk_pad),
  .clk_75mhz_i(pll_75mhz),

  .wr_fifo_empty_o(wr_fifo_empty)
);


/****************************************************************************
 * SD card + DMA core
 ****************************************************************************/
wire sd_power, sd_busy, sd_cmd, sd_cmd_oe, dmawbs_ack_o;
wire [7:0] sd_clk;
wire [31:0] sdwbs1_dat, dmawbs_dat;
wire [15:0] sdwbs1_dat_o = 
  cpuwbm2_adr_o[1] ? sdwbs1_dat[31:16] : sdwbs1_dat[15:0];
wire [31:0] sdwbs2_dat_o;
wire [15:0] dmawbs_dat_o =
  cpuwbm2_adr_o[1] ? dmawbs_dat[31:16] : dmawbs_dat[15:0];
wire sdwbs1_ack, sdwbs2_ack;
reg sdwbs1_en, sdwbs2_en, dmawbs_en;
wire [31:0] dmawbm2_adr_o, dmawbm2_dat_o;
wire dmawbm2_cyc_o, dmawbm2_stb_o, dmawbm2_we_o;
wire [3:0] dmawbm2_sel_o;
wire [2:0] dmawbm2_cti_o;
wire [1:0] dmawbm2_bte_o;
wire [3:0] sd_dat, sd_dat_oe;
assign sd_cmd_pad = sd_cmd_oe ? sd_cmd : 1'bz;
assign sd_d0_pad = sd_dat_oe[0] ? sd_dat[0] : 1'bz;
assign sd_d1_pad = sd_dat_oe[1] ? sd_dat[1] : 1'bz;
assign sd_d2_pad = sd_dat_oe[2] ? sd_dat[2] : 1'bz;
assign sd_d3_pad = sd_dat_oe[3] ? sd_dat[3] : 1'bz;
assign sd_powern_pad = un_reset ? !sd_power : 1'b0;
assign sd_clk0_pad = sd_clk[0];
assign sd_clk1_pad = sd_clk[1];
//assign act_ledn_pad = ~((cpuwbm2_cyc_o && cpuwbm2_stb_o && sdwbs1_en) ||
//  (dmawbm2_cyc_o && dmawbm2_stb_o));
assign act_ledn_pad = wr_fifo_empty;
ts_sdcore #(.multicard(1'b1)) sdcardcore(
  .wb_rst_i(wb_rst),
  .wb_clk_i(pll_100mhz & sdcard_opt),

  .wb1_adr_i(cpuwbm2_adr_o),
  .wb1_cyc_i(cpuwbm2_cyc_o && sdwbs1_en),
  .wb1_stb_i(cpuwbm2_stb_o && sdwbs1_en),
  .wb1_dat_o(sdwbs1_dat),
  .wb1_we_i(cpuwbm2_we_o),
  .wb1_dat_i({cpuwbm2_dat_o, cpuwbm2_dat_o}),
  .wb1_ack_o(sdwbs1_ack_o),
  .wb1_sel_i(cpuwbm2_32_sel_o),
  .wb1_bte_i(2'b00),
  .wb1_cti_i(3'b000),

  /* Connection to DMA core */
  .wb2_adr_i(32'h4 /* dmawbm2_adr_o */),
  .wb2_cyc_i(dmawbm2_cyc_o /* && sdwbs2_en */),
  .wb2_stb_i(dmawbm2_stb_o /* && sdwbs2_en */),
  .wb2_dat_o(sdwbs2_dat_o),
  .wb2_we_i(dmawbm2_we_o),
  .wb2_dat_i(dmawbm2_dat_o),
  .wb2_ack_o(sdwbs2_ack_o),
  .wb2_sel_i(4'b0011 /* dmawbm2_sel_o */),
  .wb2_bte_i(dmawbm2_bte_o),
  .wb2_cti_i(dmawbm2_cti_o),

  .sd_cmd_i(sd_cmd_pad),
  .sd_cmd_o(sd_cmd),
  .sd_cmd_oe_o(sd_cmd_oe),
  .sd_dat_i({sd_d3_pad, sd_d2_pad, sd_d1_pad, sd_d0_pad}),
  .sd_dat_o(sd_dat),
  .sd_dat_oe_o(sd_dat_oe),
  .sd_clk_o(sd_clk),

  .sd_wprot_i(8'b00000000),
  .sd_detect_i(8'b00000011),
  .sd_power_o(sd_power),
  .sd_busy_o(sd_busy)
);

wire dmawbm1_dack, dmawbm2_dack;
wb_dma dmacore(
  .wb_clk_i(pll_100mhz),
  .wb_rst_i(wb_rst),

  /* wbs gets connected to 100Mhz CPU wishbone */
  .wbs_cyc_i(cpuwbm2_cyc_o && dmawbs_en),
  .wbs_stb_i(cpuwbm2_stb_o && dmawbs_en),
  .wbs_we_i(cpuwbm2_we_o),
  .wbs_adr_i(cpuwbm2_adr_o),
  .wbs_dat_o(dmawbs_dat),
  .wbs_ack_o(dmawbs_ack_o),
  .wbs_dat_i({cpuwbm2_dat_o, cpuwbm2_dat_o}),
  .wbs_sel_i(cpuwbm2_32_sel_o),

  /* wbm1 gets connected to the SMC burst fifo wbs */
  .wbm1_cyc_o(cpuwbs_cyc_i),
  .wbm1_stb_o(cpuwbs_stb_i),
  .wbm1_adr_o(cpuwbs_adr_i),
  .wbm1_dat_o(cpuwbs_dat_i),
  .wbm1_dat_i(cpuwbs_dat_o),
  .wbm1_we_o(cpuwbs_we_i),
  .wbm1_ack_i(cpuwbs_ack_o),
  .wbm1_cti_o(cpuwbs_cti_i),
  .wbm1_bte_o(cpuwbs_bte_i),
  .wbm1_drq_i(1'b1), /* SMC fifo is always ready */
  .wbm1_dack_o(dmawbm1_dack),

  /* wbm2 gets connected to the SD card core */
  .wbm2_cyc_o(dmawbm2_cyc_o),
  .wbm2_stb_o(dmawbm2_stb_o),
  .wbm2_adr_o(dmawbm2_adr_o),
  .wbm2_dat_o(dmawbm2_dat_o),
  .wbm2_dat_i(sdwbs2_dat_o[15:0]),
  .wbm2_we_o(dmawbm2_we_o),
  .wbm2_ack_i(sdwbs2_ack_o),
  .wbm2_cti_o(dmawbm2_cti_o),
  .wbm2_bte_o(dmawbm2_bte_o),
  .wbm2_drq_i(1'b1),
  .wbm2_dack_o(dmawbm2_dack)
);


/****************************************************************************
 * SYSCON
 ****************************************************************************/

wire en_can, en_bbclk, en_adc, en_touch;
wire [5:0] en_txen;
wire [15:0] bus_config;

reg [15:0] fpga_irqs;

wire green_led, red_led;
assign green_led_pad = green_led ? 1'b0 : 1'b1;
assign red_led_pad = red_led ? 1'b0 : 1'b1;
reg sd_busy_q, dmawbm1_dack_q, dmawbm2_dack_q;
always @(posedge wb_clk) begin
  sd_busy_q <= sd_busy;
  dmawbm1_dack_q <= dmawbm1_dack;
  dmawbm2_dack_q <= dmawbm2_dack;
end
reg scwbs_en;
wire scwbs_ack_o, evgpio_irq;
wire [15:0] scwbs_dat_o;
syscon sysconcore(
  .wb_clk_i(wb_clk),
  .wb_rst_i(wb_rst),

  .wb_cyc_i(cpuwbm_cyc_o && scwbs_en),
  .wb_stb_i(cpuwbm_stb_o && scwbs_en),
  .wb_we_i(cpuwbm_we_o),
  .wb_adr_i(cpuwbm_adr_o),
  .wb_dat_i(cpuwbm_dat_o),
  .wb_sel_i(cpuwbm_sel_o),
  .wb_dat_o(scwbs_dat_o),
  .wb_ack_o(scwbs_ack_o),

  .dio_o(dio),
  .dio_oe_o(dio_oe),
  .dio_i(dio_in),
  .usec_i(usecs),
  .mode1_i(mode1),
  .mode2_i(mode2),

  .bus_config_o(bus_config),
  .en_can_o(en_can),
  .en_bbclk_o(en_bbclk),
  .en_txen_o(en_txen),
  .en_touch_o(en_touch),
  .red_led_o(red_led),
  .green_led_o(green_led),
  .reboot_o(reboot),
  .evgpio_irq_o(evgpio_irq),
  .irqs_i(fpga_irqs),
  .irq_o(smc_irq),
  .sd_busy_i(sd_busy_q),
  .dma_idle_i(dmawbm1_dack_q && dmawbm2_dack_q)
);

/****************************************************************************
 * 16 KByte WISHBONE blockram 
 ****************************************************************************/

wire [31:0] sramwbs1_dat;
wire [15:0] sramwbs1_dat_o = 
  cpuwbm_adr_o[1] ? sramwbs1_dat[31:16] : sramwbs1_dat[15:0];
wire [3:0] cpuwbm32_sel_o =
  cpuwbm_adr_o[1] ? {cpuwbm_sel_o, 2'b00} : {2'b00, cpuwbm_sel_o};
wire sramwbs1_ack_o;
reg sramwbs1_en;
wire xuartwbm_cyc_o, xuartwbm_stb_o, xuartwbm_we_o, xuartwbm_ack_i;
wire [15:0] xuartwbm_adr_o, xuartwbm_dat_o, xuartwbm_dat_i;
wire [31:0] xuartwbm_dat32_i;
wire [3:0] xuartwbm_sel_o;
assign xuartwbm_dat_i = xuartwbm_adr_o[1] ?
  xuartwbm_dat32_i[31:16] : xuartwbm_dat32_i[15:0];
wire sramwbm1_cyc, sramwbm1_stb, sramwbm1_ack;
wire [31:0] sramwbm1_dat;
wb_readbuf srambuf(
  .wb_clk_i(wb_clk & blockram_opt),
  .wb_rst_i(wb_rst | ~blockram_opt),

  .wbs_cyc_i(cpuwbm_cyc_o && sramwbs1_en),
  .wbs_stb_i(cpuwbm_stb_o && sramwbs1_en),
  .wbs_we_i(cpuwbm_we_o),
  .wbs_dat_o(sramwbs1_dat),
  .wbs_ack_o(sramwbs1_ack_o),

  .wbm_cyc_o(sramwbm1_cyc),
  .wbm_stb_o(sramwbm1_stb),
  .wbm_ack_i(sramwbm1_ack),
  .wbm_dat_i(sramwbm1_dat)
);
blockram sramcore(
  .wb_clk_i(wb_clk & blockram_opt),
  .wb_rst_i(wb_rst | ~blockram_opt),

  .wb1_adr_i(cpuwbm_adr_o),
  .wb1_cyc_i(sramwbm1_cyc),
  .wb1_stb_i(sramwbm1_stb),
  .wb1_we_i(cpuwbm_we_o), 
  .wb1_sel_i(cpuwbm32_sel_o),
  .wb1_dat_i({cpuwbm_dat_o, cpuwbm_dat_o}),
  .wb1_dat_o(sramwbm1_dat),
  .wb1_ack_o(sramwbm1_ack),

  .wb2_cyc_i(xuartwbm_cyc_o),
  .wb2_stb_i(xuartwbm_stb_o),
  .wb2_adr_i(xuartwbm_adr_o),
  .wb2_dat_i({xuartwbm_dat_o, xuartwbm_dat_o}),
  .wb2_sel_i(xuartwbm_sel_o),
  .wb2_we_i(xuartwbm_we_o),
  .wb2_ack_o(xuartwbm_ack_i),
  .wb2_dat_o(xuartwbm_dat32_i)

);

/****************************************************************************
 * XUART core
 ****************************************************************************/

wire xuartwbs_ack;
wire xuartwbs_ack_o = xuartwbs_ack | !xuart_opt;
reg xuwbs_en;
wire [15:0] xuartwbs_dat_o;
reg [7:0] xuart_rx;
wire [7:0] xuart_active;
wire [7:0] xuart_cts = 8'hff;
wire [7:0] xuart_txen, xuart_tx;
reg [7:0] xuart_txen_q, xuart_tx_q;
reg [7:0] xuart_rx_q, xuart_cts_q;
wire xuart_irq;
wire xwbm_cyc_o, xwbm_stb_o, xwbm_we_o, xwbm_ack_i;
wire [15:0] xwbm_adr_o, xwbm_dat_o, xwbm_dat_i;
wire [3:0] xwbm_sel32_o = xwbm_adr_o[1] ? 4'b1100 : 4'b0011;
wire [31:0] xwbm_dat32_i;
wire [3:0] xwbm_sel_o;
always @(posedge wb_clk or posedge wb_rst) begin
  if (wb_rst) begin
    xuart_rx_q <= 8'hff;
    xuart_cts_q <= 8'hff;
    xuart_tx_q <= 8'hff;
    xuart_txen_q <= 8'hff;
  end else begin
    xuart_rx_q <= xuart_rx;
    xuart_cts_q <= xuart_cts;
    xuart_tx_q <= xuart_tx;
    xuart_txen_q <= xuart_txen;
  end
end
wb_buf xubuf(
  .wb_clk_i(wb_clk),
  .wb_rst_i(wb_rst),

  .wbs_cyc_i(xwbm_cyc_o),
  .wbs_stb_i(xwbm_stb_o),
  .wbs_we_i(xwbm_we_o),
  .wbs_dat_o(xwbm_dat_i),
  .wbs_dat_i(xwbm_dat_o),
  .wbs_ack_o(xwbm_ack_i),
  .wbs_adr_i(xwbm_adr_o),
  .wbs_sel_i(xwbm_sel32_o),

  .wbm_adr_o(xuartwbm_adr_o),
  .wbm_cyc_o(xuartwbm_cyc_o),
  .wbm_stb_o(xuartwbm_stb_o),
  .wbm_we_o(xuartwbm_we_o),
  .wbm_dat_o(xuartwbm_dat_o),
  .wbm_ack_i(xuartwbm_ack_i),
  .wbm_sel_o(xuartwbm_sel_o),
  .wbm_dat_i(xuartwbm_dat_i)
);
ts_xuart tsxuartcore(
  .wb_clk_i(wb_clk & xuart_opt),
  .wb_rst_i(wb_rst | !xuart_opt),

  .wbm_cyc_o(xwbm_cyc_o),
  .wbm_stb_o(xwbm_stb_o),
  .wbm_we_o(xwbm_we_o),
  .wbm_ack_i(xwbm_ack_i),
  .wbm_adr_o(xwbm_adr_o),
  .wbm_dat_i(xwbm_dat_i),
  .wbm_dat_o(xwbm_dat_o),

  .wbs_cyc_i(cpuwbm_cyc_o && xuwbs_en),
  .wbs_stb_i(cpuwbm_stb_o && xuwbs_en),
  .wbs_we_i(cpuwbm_we_o),
  .wbs_adr_i(cpuwbm_adr_o),
  .wbs_dat_i(cpuwbm_dat_o),
  .wbs_dat_o(xuartwbs_dat_o),
  .wbs_ack_o(xuartwbs_ack),

  .baudclk_i(pll_100mhz),
  .cts_i(xuart_cts_q),
  .rx_i(xuart_rx_q),
  .tx_o(xuart_tx),
  .txen_o(xuart_txen),
  .active_o(xuart_active),
  .irq_o(xuart_irq),
  .uart_disable_i(can_opt ? 8'hf0 : 8'h80)
  //.uart_disable_i(8'h80) // 8K FPGA can hold all UARTs
);


/****************************************************************************
 * CAN core
 ****************************************************************************/

reg [4:0] seq_24mhz;
reg clk_24mhz;
always @(posedge pll_100mhz or posedge wb_rst) begin
  if (wb_rst) begin
    clk_24mhz <= 1'b1;
    seq_24mhz <= 5'd0;
  end else begin
    seq_24mhz <= seq_24mhz + 1'b1;
    case (seq_24mhz)
    1, 3, 5, 7, 9, 11, 13, 15, 17, 19, 21: clk_24mhz <= !clk_24mhz;
    24: begin
      clk_24mhz <= !clk_24mhz;
      seq_24mhz <= 5'd0;
    end
    endcase
  end
end

reg canwbs_en;
wire canwbs_ack;
wire canwbs_ack_o = canwbs_ack | !can_opt;
wire can_tx, can_irqn;
wire [7:0] canwbs_dat_i;
assign canwbs_dat_i = cpuwbm_sel_o[0] ? cpuwbm_dat_o[7:0] : cpuwbm_dat_o[15:8];
wire [7:0] canwbs_dat8_o;
wire [15:0] canwbs_dat_o = {canwbs_dat8_o, canwbs_dat8_o};
wire [7:0] canwbs_adr_i;
assign canwbs_adr_i[7:1] = cpuwbm_adr_o[7:1];
assign canwbs_adr_i[0] = !cpuwbm_sel_o[0];

can_top cancore(
  .wb_clk_i(wb_clk && can_opt),
  .wb_rst_i(wb_rst | !can_opt),
  .wb_cyc_i(cpuwbm_cyc_o && canwbs_en),
  .wb_stb_i(cpuwbm_stb_o && canwbs_en),
  .wb_we_i(cpuwbm_we_o),
  .wb_dat_i(canwbs_dat_i),
  .wb_dat_o(canwbs_dat8_o),
  .wb_adr_i(canwbs_adr_i),
  .wb_ack_o(canwbs_ack),

  .clk_i(clk_24mhz && can_opt),
  .rx_i(dio_in[6]),
  .tx_o(can_tx),
  .irq_on(can_irqn)
  //.bus_off_on()
);


/****************************************************************************
 * DIO multiplexing
 ****************************************************************************/

reg [54:0] pad_o, pad_oe;

genvar x;
generate
for (x = 0; x < 55; x = x + 1) begin: dios
  BB dio_bb(.I(pad_o[x]), .T(!pad_oe[x]), .B(dio_pad[x]), .O(dio_in[x]));
end
endgenerate

always @(*) begin
  pad_o = dio;
  pad_oe = dio_oe;

  fpga_irqs = 16'h0000;
  /* irq0: XUART
   * irq1: CAN
   * irq2: 2nd CAN
   * irq3: PC/104 irq 5
   * irq4: PC/104 irq 6
   * irq5: PC/104 irq 7
   * irq6: EVGPIO irq
   * irq7-15 reserved
   * These 16 irqs will be remapped to Linux irqs by the kernel.
   */

  fpga_irqs[1] = !can_irqn;

  if (xuart_active[7:0] != 8'h00) fpga_irqs[0] = xuart_irq;
  xuart_rx = 8'hff;

  fpga_irqs[6] = evgpio_irq;

  /* XUARTS take over DIO pins when they are activiated */
  xuart_rx[0] = dio_in[27];
  if (xuart_active[0]) begin
    pad_oe[30] = 1'b1;
    pad_o[30] = xuart_tx[0];
    pad_oe[27] = 1'b0;
  end
  xuart_rx[1] = dio_in[2];
  if (xuart_active[1]) begin
    pad_oe[1] = 1'b1;
    pad_o[1] = xuart_tx[1];
    pad_oe[2] = 1'b0;
    pad_oe[9] = 1'b1;
    pad_o[9] = xuart_txen[1];
  end
  xuart_rx[2] = dio_in[4];
  if (xuart_active[2]) begin
    pad_oe[3] = 1'b1;
    pad_o[3] = xuart_tx[2];
    pad_oe[4] = 1'b0;
    pad_o[22] = xuart_txen[2];
    pad_oe[22] = 1'b1;
  end
  xuart_rx[3] = dio_in[6];
  if (can_opt && en_can) begin
    pad_oe[5] = 1'b1;
    pad_o[5] = can_tx;
    pad_oe[6] = 1'b0;
  end else if (xuart_active[3]) begin
    pad_oe[5] = 1'b1;
    pad_o[5] = xuart_tx[3];
    pad_oe[6] = 1'b0;
  end
  xuart_rx[4] = dio_in[8];
  if (xuart_active[4]) begin
    pad_o[7] = xuart_tx[4];
    pad_oe[7] = 1'b1;
    pad_oe[8] = 1'b0;
  end 
  xuart_rx[5] = dio_in[13];
  if (xuart_active[5]) begin
    pad_oe[12] = 1'b1;
    pad_o[12] = xuart_tx[5];
    pad_oe[13] = 1'b0;
    pad_o[10] = xuart_txen[5];
    pad_oe[10] = 1'b1;
  end
  xuart_rx[6] = dio_in[21];
  if (xuart_active[6]) begin
    pad_o[14] = xuart_tx[6];
    pad_oe[14] = 1'b1;
    pad_oe[21] = 1'b0;
    pad_o[11] = xuart_txen[6];
    pad_oe[11] = 1'b1;
  end
  xuart_rx[7] = dio_in[16];
  if (xuart_active[7]) begin
    pad_o[15] = xuart_tx[7];
    pad_oe[15] = 1'b1;
    pad_oe[16] = 1'b0;
  end


end

/****************************************************************************
 * Master address decode
 ****************************************************************************/

/*

64 KB FPGA address space:
start   stop    usage
-----   ----    -----
0x8000  0xFFFF  external bus access
0x5400          XUART IO registers
0x5000          was touchscreen core
0x4D00  0x4DFF  2nd CAN core
0x4C00  0x4CFF  CAN core
0x4800          was SPI core
0x4400          ADC core
0x4000          syscon registers
0x0000  0x3FFF  16KB blockram access

*/
always @(*) begin
  
  sramwbs1_en = ts4710_decodes[0];
  xuwbs_en = ts4710_decodes[2];
  canwbs_en = ts4710_decodes[4];
  scwbs_en = ts4710_decodes[7];
  sdwbs1_en = ts4710_decodes_100mhz[8];
  dmawbs_en = ts4710_decodes_100mhz[9];
  sdwbs2_en = 1'b0;

  cpuwbm_dat_i = 16'hxxxx;
  cpuwbm_ack_i = cpuwbm_cyc_o && cpuwbm_stb_o;

  cpuwbm2_dat_i = 16'hxxxxx;
  cpuwbm2_ack_i = cpuwbm2_cyc_o && cpuwbm2_stb_o;
  cpuwbm2_en = ts4710_decodes[8] | ts4710_decodes[9];

  if (ts4710_decodes[0]) begin
    cpuwbm_dat_i = sramwbs1_dat_o;
    cpuwbm_ack_i = sramwbs1_ack_o;
  end else if (ts4710_decodes[2]) begin
    cpuwbm_dat_i = xuartwbs_dat_o;
    cpuwbm_ack_i = xuartwbs_ack_o;
  end else if (ts4710_decodes[4]) begin
    cpuwbm_dat_i = canwbs_dat_o;
    cpuwbm_ack_i = canwbs_ack_o;
  end else if (ts4710_decodes[7]) begin
    cpuwbm_dat_i = scwbs_dat_o;
    cpuwbm_ack_i = scwbs_ack_o;
  end 

  if (ts4710_decodes_100mhz[8]) begin
    cpuwbm2_dat_i = sdwbs1_dat_o;
    cpuwbm2_ack_i = sdwbs1_ack_o;
  end else if (ts4710_decodes_100mhz[9]) begin
    cpuwbm2_dat_i = dmawbs_dat_o;
    cpuwbm2_ack_i = dmawbs_ack_o;
  end
end

endmodule




