/* Copyright 2007-2008, Unpublished Work of Technologic Systems
 * All Rights Reserved.
 *
 * THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
 * PROPRIETARY AND TRADE SECRET INFORMATION OF TECHNOLOGIC SYSTEMS.
 * ACCESS TO THIS WORK IS RESTRICTED TO (I) TECHNOLOGIC SYSTEMS 
 * EMPLOYEES WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE
 * OF THEIR ASSIGNMENTS AND (II) ENTITIES OTHER THAN TECHNOLOGIC
 * SYSTEMS WHO HAVE ENTERED INTO APPROPRIATE LICENSE AGREEMENTS.  NO
 * PART OF THIS WORK MAY BE USED, PRACTICED, PERFORMED, COPIED, 
 * DISTRIBUTED, REVISED, MODIFIED, TRANSLATED, ABRIDGED, CONDENSED, 
 * EXPANDED, COLLECTED, COMPILED, LINKED, RECAST, TRANSFORMED, ADAPTED
 * IN ANY FORM OR BY ANY MEANS, MANUAL, MECHANICAL, CHEMICAL, 
 * ELECTRICAL, ELECTRONIC, OPTICAL, BIOLOGICAL, OR OTHERWISE WITHOUT
 * THE PRIOR WRITTEN PERMISSION AND CONSENT OF TECHNOLOGIC SYSTEMS.
 * ANY USE OR EXPLOITATION OF THIS WORK WITHOUT THE PRIOR WRITTEN
 * CONSENT OF TECHNOLOGIC SYSTEMS COULD SUBJECT THE PERPETRATOR TO
 * CRIMINAL AND CIVIL LIABILITY.
 */
/* Timing requirements @50Mhz (20ns bus cycle) SD speed (100Mhz wishbone)
 *
 * Output: (SD has 2ns hold and 6ns setup requirement)
 * * min tCO data - min tCO clk + 10ns > 2ns (SD hold margin)
 * * max tCO data - max tCO clk < 4ns (SD setup margin)
 *
 * Input: (SD has 14ns output delay and guarantees 2.5ns hold)
 * * max tCO clock + 14ns + tSU data < 20ns (FPGA setup margin)
 * * tHOLD data - min tCO clock < 2.5ns (FPGA hold margin)
 *
 * Timing requirements are hard to properly constrain because of
 * the relationship between tCO and tSU and the skew from the clock
 * source internally to the FPGA.  Best results are had by making the
 * clock outputs and data inputs as fast as possible and data outputs
 * as slow as possible to ensure adequate hold time.  PCB trace skew
 * helps FPGA hold margin, hurts FPGA setup margin.  SD hold margin can
 * be gained by adding combinational delays before the data outputs--
 * output registers in the I/O ring should not be used for data out.
 *
 * Prosoft-8000 SD margins as of 8/19/2008 (10 pF)
 * SD hold margin: .783ns
 * SD setup margin: 3.874ns
 * FPGA setup margin: .450ns
 * FPGA hold margin: 3.762ns
 */

module ts_sdcore(
  wb_clk_i,
  wb_rst_i,

  wb1_cyc_i,
  wb1_cti_i,
  wb1_bte_i,
  wb1_stb_i,
  wb1_we_i,
  wb1_adr_i,
  wb1_dat_i,
  wb1_sel_i,
  wb1_dat_o,
  wb1_ack_o,

  wb2_cyc_i,
  wb2_stb_i,
  wb2_cti_i,
  wb2_bte_i,
  wb2_we_i,
  wb2_adr_i,
  wb2_dat_i,
  wb2_sel_i,
  wb2_dat_o,
  wb2_ack_o,

  wb3_cyc_i,
  wb3_stb_i,
  wb3_cti_i,
  wb3_bte_i,
  wb3_we_i,
  wb3_adr_i,
  wb3_dat_i,
  wb3_sel_i,
  wb3_dat_o,
  wb3_ack_o,

  wb4_cyc_i,
  wb4_stb_i,
  wb4_cti_i,
  wb4_bte_i,
  wb4_we_i,
  wb4_adr_i,
  wb4_dat_i,
  wb4_sel_i,
  wb4_dat_o,
  wb4_ack_o,

  sd_clk_o,
  sd_cmd_o,
  sd_cmd_i,
  sd_cmd_oe_o,
  sd_dat_o,
  sd_dat_i,
  sd_dat_oe_o,
  sd_power_o,
  sd_detect_i,
  sd_wprot_i,
  sd_busy_o

);

input wb_clk_i, wb_rst_i;

input wb1_cyc_i, wb1_stb_i, wb1_we_i;
input [3:0] wb1_sel_i;
input [31:0] wb1_adr_i, wb1_dat_i;
output [31:0] wb1_dat_o;
output wb1_ack_o;
input [2:0] wb1_cti_i;
input [1:0] wb1_bte_i;

input wb2_cyc_i, wb2_stb_i, wb2_we_i;
input [3:0] wb2_sel_i;
input [31:0] wb2_adr_i, wb2_dat_i;
output [31:0] wb2_dat_o;
output wb2_ack_o;
input [2:0] wb2_cti_i;
input [1:0] wb2_bte_i;

input wb3_cyc_i, wb3_stb_i, wb3_we_i;
input [3:0] wb3_sel_i;
input [31:0] wb3_adr_i, wb3_dat_i;
output [31:0] wb3_dat_o;
output wb3_ack_o;
input [2:0] wb3_cti_i;
input [1:0] wb3_bte_i;

input wb4_cyc_i, wb4_stb_i, wb4_we_i;
input [3:0] wb4_sel_i;
input [31:0] wb4_adr_i, wb4_dat_i;
output [31:0] wb4_dat_o;
output wb4_ack_o;
input [2:0] wb4_cti_i;
input [1:0] wb4_bte_i;

output [7:0] sd_clk_o;
output sd_cmd_o, sd_cmd_oe_o, sd_power_o;
output [3:0] sd_dat_o;
output [3:0] sd_dat_oe_o;
input sd_cmd_i;
input [7:0] sd_wprot_i, sd_detect_i;
input [3:0] sd_dat_i;
output sd_busy_o;

parameter multicard = 1'b1;

wire [31:0] wb_adr_i;
wire [31:0] wb_dat_i;
wire [3:0] wb_sel_i;
wire [31:0] wb_dat_o;
wire [2:0] wb_cti_i;
wire [1:0] wb_bte_i;
wire wb_ack_o, wb_cyc_i, wb_stb_i, wb_we_i;
wb_arbiter arbitercore(
  .wb_clk_i(wb_clk_i),

  .wbowner_dat_o(wb_dat_i),
  .wbowner_dat_i(wb_dat_o),
  .wbowner_adr_o(wb_adr_i),
  .wbowner_cyc_o(wb_cyc_i),
  .wbowner_stb_o(wb_stb_i),
  .wbowner_we_o(wb_we_i),
  .wbowner_ack_i(wb_ack_o),
  .wbowner_sel_o(wb_sel_i),
  .wbowner_cti_o(wb_cti_i),
  .wbowner_bte_o(wb_bte_i),

  .wb1_adr_i(wb1_adr_i),
  .wb1_dat_i(wb1_dat_i),
  .wb1_dat_o(wb1_dat_o),
  .wb1_cyc_i(wb1_cyc_i),
  .wb1_stb_i(wb1_stb_i),
  .wb1_we_i(wb1_we_i),
  .wb1_ack_o(wb1_ack_o),
  .wb1_sel_i(wb1_sel_i),
  .wb1_cti_i(wb1_cti_i),
  .wb1_bte_i(wb1_bte_i),

  .wb2_adr_i(wb2_adr_i),
  .wb2_dat_i(wb2_dat_i),
  .wb2_dat_o(wb2_dat_o),
  .wb2_cyc_i(wb2_cyc_i),
  .wb2_stb_i(wb2_stb_i),
  .wb2_we_i(wb2_we_i),
  .wb2_ack_o(wb2_ack_o),
  .wb2_sel_i(wb2_sel_i),
  .wb2_cti_i(wb2_cti_i),
  .wb2_bte_i(wb2_bte_i),

  .wb3_adr_i(wb3_adr_i),
  .wb3_dat_i(wb3_dat_i),
  .wb3_dat_o(wb3_dat_o),
  .wb3_cyc_i(wb3_cyc_i),
  .wb3_stb_i(wb3_stb_i),
  .wb3_we_i(wb3_we_i),
  .wb3_ack_o(wb3_ack_o),
  .wb3_sel_i(wb3_sel_i),
  .wb3_cti_i(wb3_cti_i),
  .wb3_bte_i(wb3_bte_i),

  .wb4_adr_i(wb4_adr_i),
  .wb4_dat_i(wb4_dat_i),
  .wb4_dat_o(wb4_dat_o),
  .wb4_cyc_i(wb4_cyc_i),
  .wb4_stb_i(wb4_stb_i),
  .wb4_we_i(wb4_we_i),
  .wb4_ack_o(wb4_ack_o),
  .wb4_sel_i(wb4_sel_i),
  .wb4_cti_i(wb4_cti_i),
  .wb4_bte_i(wb4_bte_i)
);

reg [31:0] wb_dat;
reg [3:0] sd_ldat;
reg [3:0] sd_dat /* synthesis syn_useioff=0 */;
reg [3:0] sd_dat_oe;
reg [3:0] state;
reg sd_lcmd;
reg sd_cmd /* synthesis syn_useioff=0 */;
reg sd_cmd_oe, sd_power, wb_ack;
reg [7:0] sd_clk;
reg [7:0] sd_oclk /* synthesis syn_useioff=1 */;
reg [1:0] sd_oe;
reg [2:0] nibble;
assign sd_clk_o = sd_oclk;
assign sd_dat_o = sd_dat;
assign sd_dat_oe_o = sd_dat_oe;
assign sd_cmd_o = sd_cmd;
assign sd_cmd_oe_o = sd_cmd_oe;
assign wb_ack_o = wb_ack;
assign wb_dat_o = wb_dat;
assign sd_power_o = sd_power;
assign sd_busy_o = (state != 4'd0);

always @(sd_oe) begin
  case (sd_oe) /* synthesis full_case, parallel_case */
  2'd0: {sd_cmd_oe, sd_dat_oe} = 5'b11111;
  2'd1: {sd_cmd_oe, sd_dat_oe} = 5'b00001;
  2'd2: {sd_cmd_oe, sd_dat_oe} = 5'b10000;
  2'd3: {sd_cmd_oe, sd_dat_oe} = 5'b00000;
  endcase
end

function [6:0] clk_select;
input [2:0] sel;
begin
  case (sel) /* synthesis full_case, parallel_case */
  3'd0: clk_select = 7'b0000001;
  3'd1: clk_select = 7'b0000010;
  3'd2: clk_select = 7'b0000100;
  3'd3: clk_select = 7'b0001000;
  3'd4: clk_select = 7'b0010000;
  3'd5: clk_select = 7'b0100000;
  3'd6: clk_select = 7'b1000000;
  3'd7: clk_select = 7'b1111111;
  endcase
end
endfunction 

wire stall;
reg crc_en, crc_rst, crc_shift;
wire [15:0] crc0, crc1, crc2, crc3;
wire [6:0] cmdcrc;
reg [3:0] sd_dat_i_q /* synthesis syn_useioff=1 */;
reg sd_cmd_i_q /* synthesis syn_useioff=1 */;
crc crc16core0(
  .dat_i(sd_ldat[0]),
  .clk_i(wb_clk_i),
  .clken_i(crc_en && !stall),
  .shift_i(crc_shift),
  .rst_i(crc_rst),
  .crc16_o(crc0)
);
crc crc16core1(
  .dat_i(sd_ldat[1]),
  .clk_i(wb_clk_i),
  .clken_i(crc_en && !stall),
  .shift_i(crc_shift),
  .rst_i(crc_rst),
  .crc16_o(crc1)
);
crc crc16core2(
  .dat_i(sd_ldat[2]),
  .clk_i(wb_clk_i),
  .clken_i(crc_en && !stall),
  .shift_i(crc_shift),
  .rst_i(crc_rst),
  .crc16_o(crc2)
);
crc crc16core3(
  .dat_i(sd_ldat[3]),
  .clk_i(wb_clk_i),
  .clken_i(crc_en && !stall),
  .shift_i(crc_shift),
  .rst_i(crc_rst),
  .crc16_o(crc3)
);
crc crc7core(
  .dat_i(sd_lcmd),
  .clk_i(wb_clk_i),
  .clken_i(crc_en && !stall),
  .shift_i(crc_shift),
  .rst_i(crc_rst),
  .crc7_o(cmdcrc)
);

/*
 * TS-SDCORE version auto-detection, byte read at:
 * 
 * base + 0x3: bit 6 is read/writeable == 8-bit TS-SDCORE v1
 * base + 0x3: bit 6 is stuck 0 == 32-bit TS-SDCORE v2
 * base + 0x3: bit 6 is stuck 1 == 8-bit TS-SDCORE v2
 */

reg [3:0] dat;
always @(nibble or wb_dat_i) begin
  case (nibble) /* synthesis full_case, parallel_case */
  3'd1: dat = wb_dat_i[3:0];
  3'd0: dat = wb_dat_i[7:4];
  3'd3: dat = wb_dat_i[11:8];
  3'd2: dat = wb_dat_i[15:12];
  3'd5: dat = wb_dat_i[19:16];
  3'd4: dat = wb_dat_i[23:20];
  3'd7: dat = wb_dat_i[27:24];
  3'd6: dat = wb_dat_i[31:28];
  endcase
end

reg [7:0] crcack;
reg [9:0] nclks;
reg [15:0] loops;
reg [2:0] dat_hist, clksel;
reg [3:0] wb_sel;
reg magic;
reg burst, crc_err, wb_we, rdmult_en, wrmult_en;
reg sd_clk_prev, fast_en, timeout, incloop, bsywait, wb_preack;
assign stall = ((sd_clk_prev != sd_clk[clksel]) && !fast_en);
always @(posedge wb_clk_i) sd_dat_i_q <= sd_dat_i;
always @(posedge wb_clk_i) sd_cmd_i_q <= sd_cmd_i;
always @(posedge wb_clk_i or posedge wb_rst_i) begin
  if (wb_rst_i) begin
    sd_oclk <= 8'd0;
    magic <= 1'b0;
    sd_lcmd <= 1'b0;
    sd_cmd <= 1'b0;
    sd_dat <= 4'd0;
    sd_ldat <= 4'd0;
    sd_clk <= 8'd0;
    sd_clk_prev <= 1'b0;
    state <= 4'd0;
    sd_oe <= 2'd0;
    wb_ack <= 1'b0;
    wb_dat <= 32'd0;
    sd_power <= 1'b0;
    nibble <= 3'd0;
    crc_en <= 1'b0;
    crc_rst <= 1'b0;
    nclks <= 10'd0;
    crc_shift <= 1'b0;
    dat_hist <= 3'd0;
    burst <= 1'b0;
    crc_err <= 1'b0;
    wb_we <= 1'b0;
    wb_sel <= 4'd0;
    rdmult_en <= 1'b1;
    wrmult_en <= 1'b1;
    fast_en <= 1'b0;
    crcack <= 8'd0;
    timeout <= 1'b0;
    incloop <= 1'b0;
    loops <= 16'd0;
    bsywait <= 1'b0;
    clksel <= 3'd0;
    wb_preack <= 1'b0;
  end else begin
    if (wb_cyc_i && wb_stb_i)
      burst <= (wb_cti_i != 3'b111 && wb_cti_i != 3'b000);
    else if (!wb_cyc_i)
      burst <= 1'b0;

    if (wb_ack && wb_cyc_i && !wb_stb_i && burst) wb_ack <= 1'b1;
    else wb_ack <= 1'b0; 
    crc_rst <= 1'b0;
    crc_shift <= 1'b0;
    if (sd_ldat == 4'd0 && !sd_lcmd && !sd_clk[clksel] && sd_oe == 2'd0) 
      sd_power <= 1'b0;
    else sd_power <= 1'b1;
    sd_clk_prev <= sd_clk[clksel];
    if (incloop) begin
      loops <= loops + 1'b1;
      incloop <= 1'b0;
    end
    if (loops[15]) timeout <= 1'b1;

    if (!stall) case (state)
    4'd0: begin
      loops <= 16'd0;
      bsywait <= 1'b0;
      if (wb_cyc_i && wb_stb_i && (!wb_ack || burst)) begin
        wb_we <= wb_we_i;
        wb_sel <= wb_sel_i;
        case (wb_adr_i[3:2])
        2'd0: begin
          if (wb_sel_i[0] && wb_we_i) begin
            sd_dat <= wb_dat_i[3:0];
            sd_ldat <= wb_dat_i[3:0];
            sd_cmd <= wb_dat_i[4];
            sd_lcmd <= wb_dat_i[4];
            sd_clk[clksel] <= wb_dat_i[5];
            sd_oclk[clksel] <= wb_dat_i[5];
            sd_oe <= wb_dat_i[7:6];
            crc_rst <= 1'b1;
            nclks <= 10'd0;
          end
          if (wb_sel_i[1] && wb_we_i) begin
            rdmult_en <= wb_dat_i[11];
            wrmult_en <= wb_dat_i[12];
            fast_en <= wb_dat_i[13];
          end
          if (wb_sel_i[2] && wb_we_i) begin
            clksel <= wb_dat_i[18:16];
            sd_oclk[wb_dat_i[18:16]] <= sd_clk[clksel];
            sd_clk[wb_dat_i[18:16]] <= sd_clk[clksel];
          end
          if (!wb_we_i && wb_sel_i[1]) begin
            crc_err <= 1'b0;
            timeout <= 1'b0;
          end
          wb_ack <= 1'b1;
          wb_dat[7:0] <= {sd_oe, sd_clk[clksel], sd_cmd_i_q, sd_dat_i_q};
          wb_dat[14:8] <= {timeout, fast_en, wrmult_en, rdmult_en, crc_err, 
            sd_wprot_i[clksel], sd_detect_i[clksel]};
          // For detection of 32-bit TS-SDCORE v2, bit 30 of address 0 is
          // stuck 0.  In the 8-bit TS-SDCORE v1, this bit is 
          // readable/writeable (quadbit_en)
          wb_dat[30] <= 1'b0; 
        end
        2'd1: begin /* stream 4-bit read/write on sd_dat bus */
          sd_oclk[clksel] <= sd_clk[clksel] ^ 1'b1;
          sd_clk[clksel] <= sd_clk[clksel] ^ 1'b1;
          nclks <= nclks + 1'b1;
          sd_dat <= dat;
          sd_ldat <= dat;
          crc_en <= 1'b1;
          state <= 4'd1;
          if (wb_preack) wb_ack <= 1'b1;
        end
        2'd2: begin /* stream 1-bit read/write on sd_cmd */
          wb_dat <= wb_dat_i[31:0];
          state <= 4'd13;
        end
        2'd3: begin
          // For detection of 32-bit TS-SDCORE v2 vs. 8-bit TS-SDCORE v2.
          wb_dat[6] <= !magic;
          if (!wb_sel_i[0]) magic <= 1'b1;
          else magic <= 1'b0;
          wb_ack <= 1'b1;
          if (wb_we_i) begin
            state <= 4'd8;
            bsywait <= 1'b1;
          end
        end
        default: wb_ack <= 1'b1;
        endcase
      end
    end
    4'd1: begin
      crc_en <= 1'b0;
      nibble <= nibble + 1'b1;
      case (nibble)
      3'd0: begin
        wb_dat[7:4] <= sd_dat_i_q;
        wb_dat[23:20] <= sd_dat_i_q;
	if (!wb_sel[1] && fast_en && wb_we_i) wb_preack <= 1'b1;
      end
      3'd1: begin
        wb_dat[3:0] <= sd_dat_i_q;
        wb_dat[19:16] <= sd_dat_i_q;
        if (!wb_sel[1]) begin
          if (!wb_preack) wb_ack <= 1'b1;
          nibble <= 3'd0;
          wb_preack <= 1'b0;
        end
      end
      3'd2: begin
        wb_dat[15:12] <= sd_dat_i_q;
	if (!wb_sel[2] && fast_en && wb_we_i) wb_preack <= 1'b1;
      end
      3'd3: begin
        wb_dat[11:8] <= sd_dat_i_q;
        if (!wb_sel[2]) begin
          if (!wb_preack) wb_ack <= 1'b1;
          nibble <= 3'd0;
          wb_preack <= 1'b0;
        end
      end
      3'd4: begin
        wb_dat[23:20] <= sd_dat_i_q;
	if (!wb_sel[3] && fast_en && wb_we_i) wb_preack <= 1'b1;
      end
      3'd5: begin
        wb_dat[19:16] <= sd_dat_i_q;
        if (!wb_sel[3]) begin
          if (!wb_preack) wb_ack <= 1'b1;
          nibble <= 3'd0;
          wb_preack <= 1'b0;
        end
      end
      3'd6: begin
        wb_dat[31:28] <= sd_dat_i_q;
	if (fast_en && wb_we_i) wb_preack <= 1'b1;
      end
      3'd7: begin
        wb_dat[27:24] <= sd_dat_i_q;
        if (!wb_preack) wb_ack <= 1'b1;
        nibble <= 3'd0;
	wb_preack <= 1'b0;
      end
      endcase
      state <= 4'd0;
      sd_clk[clksel] <= sd_clk[clksel] ^ 1'b1; // w -> 1, r -> 0
      sd_oclk[clksel] <= sd_clk[clksel] ^ 1'b1;
      if (nclks == 10'd0) begin // rollover
        state <= 4'd2; 
      end
    end
    4'd2: begin // advance 16 clks (crc)
      sd_clk[clksel] <= sd_clk[clksel] ^ 1'b1; // w -> 0, r -> 1
      sd_oclk[clksel] <= sd_clk[clksel] ^ 1'b1;
      crc_shift <= 1'b1;
      sd_dat <= {crc3[15], crc2[15], crc1[15], crc0[15]};
      nclks <= nclks + 1'b1;
      state <= 4'd3;
    end
    4'd3: begin
//      if (!wb_we && sd_dat != sd_dat_i_q) crc_err <= 1'b1;
      sd_clk[clksel] <= sd_clk[clksel] ^ 1'b1; // w -> 1, r -> 0
      sd_oclk[clksel] <= sd_clk[clksel] ^ 1'b1;
      if (nclks[4:0] >= 10'd16) begin
        state <= 4'd4;
      end else state <= 4'd2;
    end
    4'd4: begin // end
      sd_clk[clksel] <= sd_clk[clksel] ^ 1'b1; // w-> 0, r -> 1
      sd_oclk[clksel] <= sd_clk[clksel] ^ 1'b1;
      nclks <= nclks + 1'b1;
      sd_dat <= 4'b1111;
      sd_ldat <= 4'b1111;
      if (wb_we && !timeout) state <= 4'd6; // Must step through CRC ack
      else if (rdmult_en && !timeout) state <= 4'd5; // read continues...
      else state <= 4'd11;
    end
    4'd5: begin // wait for next read block
      if (nclks == 10'd0) incloop <= 1'b1;
      sd_clk[clksel] <= sd_clk[clksel] ^ 1'b1; // w-> 1, r -> 0
      sd_oclk[clksel] <= sd_clk[clksel] ^ 1'b1;
      if (sd_dat_i_q[0] == 1'b0) begin
        state <= 4'd0; // okay to continue
        nclks <= 10'd0;
      end
      else state <= 4'd4;
    end
    4'd6: begin // CRC acknowledge
      sd_clk[clksel] <= sd_clk[clksel] ^ 1'b1; // w-> 1
      sd_oclk[clksel] <= sd_clk[clksel] ^ 1'b1;
      sd_oe <= 2'd2; // tri-state dat lines
      if (nclks[3:0] >= 10'd8) begin
        if (crcack[2:0] != 3'b010) crc_err <= 1'b1;
        if (wrmult_en) state <= 4'd8;
        else state <= 4'd11;
      end else state <= 4'd7;
    end
    4'd7: begin
      sd_clk[clksel] <= sd_clk[clksel] ^ 1'b1; // w-> 0
      sd_oclk[clksel] <= sd_clk[clksel] ^ 1'b1;
      crcack <= {crcack[6:0], sd_dat_i_q[0]};
      nclks <= nclks + 1'b1;
      state <= 4'd6;
    end
    4'd8: begin // CRC ack is complete, wait for card not busy
      sd_clk[clksel] <= sd_clk[clksel] ^ 1'b1; // w-> 0
      sd_oclk[clksel] <= sd_clk[clksel] ^ 1'b1;
      state <= 4'd9;
      dat_hist <= {dat_hist[1:0], sd_dat_i_q[0]};
      if (nclks == 10'd0) incloop <= 1'b1;
    end
    4'd9: begin
      sd_clk[clksel] <= sd_clk[clksel] ^ 1'b1; // w-> 1
      sd_oclk[clksel] <= sd_clk[clksel] ^ 1'b1; 
      nclks <= nclks + 1'b1;
      if (timeout) begin
        dat_hist <= 3'b000;
        state <= 4'd10;
      end else if (dat_hist == 3'b111) begin
        dat_hist <= 3'b000;
        sd_oe <= 2'd0; // Back to driving sd_dat
        if (bsywait) state <= 4'd0;
        else state <= 4'd10; // clock a start condition
      end else begin
        state <= 4'd8;
      end
    end
    4'd10: begin
      sd_clk[clksel] <= sd_clk[clksel] ^ 1'b1;
      sd_oclk[clksel] <= sd_clk[clksel] ^ 1'b1;
      sd_dat <= 4'd0;
      sd_ldat <= 4'd0;
      state <= 4'd11;
    end
    4'd11: begin
      sd_clk[clksel] <= sd_clk[clksel] ^ 1'b1;
      sd_oclk[clksel] <= sd_clk[clksel] ^ 1'b1;
      state <= 4'd0;
      nclks <= 10'd0;
    end
    4'd12: begin
      crc_en <= 1'b0;
      state <= 4'd13;
      wb_dat <= {wb_dat[30:0], sd_cmd_i_q};
      if (nclks[1:0] == 2'd0) case (nibble)
      3'd0: nibble <= nibble + 1'b1;
      3'd1: begin
        if (!wb_sel[1]) begin
          wb_dat[23:16] <= {wb_dat[6:0], sd_cmd_i_q};
          wb_ack <= 1'b1;
          state <= 4'd0;
          nibble <= 3'd0;
        end else nibble <= nibble + 1'b1;
      end
      3'd2: nibble <= nibble + 1'b1;
      3'd3: begin
        if (!wb_sel[2]) begin
          wb_ack <= 1'b1;
          state <= 4'd0;
          nibble <= 3'd0;
        end else nibble <= nibble + 1'b1;
      end
      3'd4: nibble <= nibble + 1'b1;
      3'd5: begin
        if (!wb_sel[3]) begin
          wb_ack <= 1'b1;
          state <= 4'd0;
          nibble <= 3'd0;
        end else nibble <= nibble + 1'b1;
      end
      3'd6: nibble <= nibble + 1'b1;
      3'd7: begin
        wb_ack <= 1'b1;
        nibble <= 3'd0;
        state <= 4'd0;
      end
      endcase
      sd_clk[clksel] <= sd_clk[clksel] ^ 1'b1; // w -> 1, r -> 0
      sd_oclk[clksel] <= sd_clk[clksel] ^ 1'b1; 
      if (wb_we && nclks[5:0] >= 6'd40) begin
        state <= 4'd14; // crc7
        wb_ack <= 1'b1;
        nibble <= 3'd0;
      end
    end
    4'd13: begin
      sd_clk[clksel] <= sd_clk[clksel] ^ 1'b1;
      sd_oclk[clksel] <= sd_clk[clksel] ^ 1'b1;
      crc_en <= 1'b1;
      nclks <= nclks + 1'b1;
      if (wb_sel[3]) sd_cmd <= wb_dat[31];
      else if (wb_sel[1]) sd_cmd <= wb_dat[15];
      else sd_cmd <= wb_dat[7];

      if (wb_sel[3]) sd_lcmd <= wb_dat[31];
      else if (wb_sel[1]) sd_lcmd <= wb_dat[15];
      else sd_lcmd <= wb_dat[7];

      state <= 4'd12;
    end
    4'd14: begin
      sd_clk[clksel] <= sd_clk[clksel] ^ 1'b1; // w -> 0, r -> 1
      sd_oclk[clksel] <= sd_clk[clksel] ^ 1'b1; 
      crc_shift <= 1'b1;
      sd_cmd <= cmdcrc[6];
      nclks <= nclks + 1'b1;
      state <= 4'd15;
    end
    4'd15: begin
      sd_clk[clksel] <= sd_clk[clksel] ^ 1'b1; // w -> 1, r -> 0
      sd_oclk[clksel] <= sd_clk[clksel] ^ 1'b1;
      if (nclks[5:0] >= 6'd48) begin
        state <= 4'd0;
      end else state <= 4'd14;
    end
    default: state <= 4'bxxxx;    
    endcase
    
    if (!multicard) clksel <= 3'd0;
  end
end

endmodule
